ifeq (,$(ShakeNMake.MAKEFILE))
# ^^^ multiple-inclusion guard, now that modules are autonomously
# including this file.
all:
SHELL:=/bin/bash

ShakeNMake.MAKEFILE := $(word $(words $(MAKEFILE_LIST)),$(MAKEFILE_LIST))
ShakeNMake.DIR := $(realpath $(dir $(ShakeNMake.MAKEFILE)))
$(ShakeNMake.MAKEFILE):# avoid breaking some deps checks if someone renames this file (been there, done that)

PACKAGE.MAKEFILE := $(firstword $(MAKEFILE_LIST))# normally either Makefile or GNUmakefile
$(PACKAGE.MAKEFILE):

########################################################################
# Maintainer's/hacker's notes:
#
# Vars names starting with ShakeNMake are mostly internal to this
# makefile and are considered "private" unless documented otherwise.
# Notable exceptions are most of the ShakeNMake.CALL entries, which
# are $(call)able functions, and ShakeNMake.EVAL entries, which are
# $(eval)able code.
#
########################################################################


########################################################################
# ShakeNMake.CALL.FIND_FILE call()able function:
# $1 = app name
# $2 = optional path ($(PATH) is the default)
ShakeNMake.CALL.FIND_FILE = $(firstword $(wildcard $(addsuffix /$(1),$(subst :, ,$(if $(2),$(2),$(PATH))))))

########################################################################
# ShakeNMake.CALL.FIND_HEADER: $(call)able function. $(1) should be
# the the name of a file to search for $(ShakeNMake.PATH.HEADERS).
ShakeNMake.PATH.HEADERS ?= $(HOME)/include /usr/local/include /usr/include
define ShakeNMake.CALL.FIND_HEADER
$(call ShakeNMake.CALL.FIND_FILE,$(1),$(2) $(ShakeNMake.PATH.HEADERS))
endef
########################################################################
# ShakeNMake.CALL.FIND_LIB: $(call)able function. $(1) should be the
# the name of a file to search for in $(ShakeNMake.PATH.LIBS).
# If $1 is multiple names, the first one matched is returned.
ShakeNMake.PATH.LIBS ?= $(HOME)/lib /usr/local/lib /usr/lib /usr/lib/x86_64-linux-gnu
define ShakeNMake.CALL.FIND_LIB
$(call ShakeNMake.CALL.FIND_FILE,$(1),$(2) $(ShakeNMake.PATH.LIBS))
endef

########################################################################
ShakeNMake.BINS.AR := $(call ShakeNMake.CALL.FIND_FILE,$(AR))
ShakeNMake.BINS.GCC := $(call ShakeNMake.CALL.FIND_FILE,gcc)


ifneq (,$(COMSPEC))
$(warning Setting ShakeNMake.SMELLS.LIKE.WINDOWS to 1)
ShakeNMake.SMELLS.LIKE.WINDOWS := 1
ShakeNMake.EXTENSIONS.DLL = .DLL# maintenance reminder: this must stay upper-case!
ShakeNMake.EXTENSIONS.EXE = .EXE# maintenance reminder: this must stay upper-case!
else
ShakeNMake.SMELLS.LIKE.WINDOWS := 0
ShakeNMake.EXTENSIONS.DLL = .so
ShakeNMake.EXTENSIONS.EXE =# no whitespace, please
endif

########################################################################
# Cleanup rules.
# To ensure proper cleanup order, define deps on clean-., like so:
# clean-.: clean-mysubdir
# clean-mysubdir: clean-prereqsubdir
# ...
SUBMAKE.INDENT := '===== '
.PHONY: clean distclean clean-. distclean-.
clean-%:
	@$(MAKE) -C $* --no-print-directory clean
clean-.:
	@-deep=0; while [[ $$deep -lt $(MAKELEVEL) ]]; do echo -n $(SUBMAKE.INDENT); deep=$$((deep+1)); done; \
	echo "Cleaning up $(CURDIR)..."; \
	touch $@.$$$$; rm -f $@.$$$$ $(CLEAN_FILES) 
clean: clean-.

distclean-%:
	@$(MAKE) -C $* --no-print-directory distclean
distclean-.:
	@-deep=0; while [[ $$deep -lt $(MAKELEVEL) ]]; do echo -n $(SUBMAKE.INDENT); deep=$$((deep+1)); done; \
	echo "Dist-cleaning $(CURDIR)..."; \
	touch $@.$$$$; rm -f $@.$$$$ $(CLEAN_FILES) $(DISTCLEAN_FILES) 
distclean: | distclean-.
# /clean/distclean
########################################################################

########################################################################
# An internal hack to enable "quiet" output. $(1) is a string which is
# shown ONLY if ShakeNMake.QUIET=1 (otherwise set -x gets called for
# the current shell process).
ShakeNMake.QUIET ?= 0
define ShakeNMake.CALL.SETX
if test x1 = "x$(ShakeNMake.QUIET)"; then echo $(1); else set -x; fi
endef
########################################################################

########################################################################
# Building a static library:
#
#   myLib.LIB.OBJECTS = foo.o bar.o
#   $(call ShakeNMake.CALL.RULES.LIBS,myLib)
#
# A target named myLib.LIB is created for building the library and
# $(myLib.LIB) holds the name of the compiled lib (typically
# "myLib.a").
#
########################################################################
# ShakeNMake.CALL.RULES.LIB creates rules to build static library
# $(1).a
define ShakeNMake.CALL.RULES.LIB
$(1).LIB = $(1).a
$(1).LIB: $$($(1).LIB)
CLEAN_FILES += $$($(1).LIB) $$($(1).LIB.OBJECTS)
$$($(1).LIB): $$($(1).LIB.OBJECTS)
	@$$(call ShakeNMake.CALL.SETX,"AR [$$@] ..."); \
		$$(ShakeNMake.BINS.AR) crs $$@ $$($(1).LIB.OBJECTS)
endef
define ShakeNMake.CALL.RULES.LIBS
$(foreach liba,$(1),$(eval $(call ShakeNMake.CALL.RULES.LIB,$(liba))))
endef
# end ShakeNMake.CALL.RULES.LIB
########################################################################

########################################################################
# ShakeNMake.CALL.RULES.DLL builds $(1)$(ShakeNMake.EXTENSIONS.DLL)
# from object files defined by $(1).DLL.OBJECTS and
# $(1).DLL.SOURCES. Flags passed on to the shared library linker:
#
#   LDFLAGS, $(1).DLL.LDFLAGS, LDADD, -shared
#   $(1).DLL.OBJECTS
#
# Also defines the var $(1).DLL, which expands to the filename of the DLL,
# (normally $(1)$(ShakeNMake.EXTENSIONS.DLL)).
define ShakeNMake.CALL.RULES.DLL
$(1).DLL = $(1)$$(ShakeNMake.EXTENSIONS.DLL)
ifneq (.DLL,$$(ShakeNMake.EXTENSIONS.DLL))
$(1).DLL: $$($(1).DLL)
endif
#$(1).DLL.OBJECTS_FROM_SOURCES := $$(addsuffix .o,$$(basename $$($(1).DLL.SOURCES)))
CLEAN_FILES += $$($(1).DLL) $$($(1).DLL.OBJECTS)
$$($(1).DLL): $$($(1).DLL.OBJECTS) $$($(1).DLL.SOURCES)
	@test x = "x$$($(1).DLL.OBJECTS)$$($(1).DLL.SOURCES)" && { \
	echo "$(1).DLL.OBJECTS and/or $(1).DLL.SOURCES are/is undefined!"; exit 1; }; \
	$$(call ShakeNMake.CALL.SETX,"LD [$$@] ..."); \
	 $$(CC) -o $$@ -shared $$(LDFLAGS) \
		$$($(1).DLL.CPPFLAGS) \
		$$($(1).DLL.SOURCES) \
		$$($(1).DLL.OBJECTS) \
		$$($(1).DLL.LDFLAGS) \
		$$($(1).DLL.CXXFLAGS)
endef
########################################################################
# $(call ShakeNMake.CALL.RULES.DLLS,[list]) calls and $(eval)s
# ShakeNMake.CALL.RULES.DLL for each entry in $(1)
define ShakeNMake.CALL.RULES.DLLS
$(foreach dll,$(1),$(eval $(call ShakeNMake.CALL.RULES.DLL,$(dll))))
endef
# end ShakeNMake.CALL.RULES.DLLS and friends
########################################################################


########################################################################
# ShakeNMake.CALL.RULES.BIN is intended to be called like so:
# $(eval $(call ShakeNMake.CALL.RULES.BIN,MyApp))
#
# It builds a binary named $(1) by running $(CC) and passing it:
#
# $(1).BIN.OBJECTS, LDFLAGS, $(1).BIN.LDFLAGS
define ShakeNMake.CALL.RULES.BIN
$(1).BIN := $(1)$(ShakeNMake.EXTENSIONS.EXE)
$(1).BIN: $$($(1).BIN)
# Many developers feel that bins should not be cleaned by 'make
# clean', but instead by distclean, but i'm not one of those
# developers. i subscribe more to the school of thought that distclean
# is for cleaning up configure-created files. That said, shake-n-make
# isn't designed to use a configure-like process, so that is probably
# moot here and we probably (maybe?) should clean up bins only in
# distclean. As always: hack it to suit your preference:
CLEAN_FILES += $$($(1).BIN) $$($(1).BIN.OBJECTS)
$$($(1).BIN): $$($(1).BIN.OBJECTS)
	@test x = "x$$($(1).BIN.OBJECTS)" && { \
	echo "$(1).BIN.OBJECTS is undefined!"; exit 1; }; \
	$(call ShakeNMake.CALL.SETX,"CC [$$@] ..."); \
	$$(CC) -o $$@ \
		$$($(1).BIN.OBJECTS) \
		$$(LDFLAGS) $$($(1).BIN.LDFLAGS)
endef
########################################################################
# $(call ShakeNMake.CALL.RULES.BINS,[list]) calls and $(eval)s
# ShakeNMake.CALL.RULES.BIN for each entry in $(1)
define ShakeNMake.CALL.RULES.BINS
$(foreach bin,$(1),$(eval $(call ShakeNMake.CALL.RULES.BIN,$(bin))))
endef
# end ShakeNMake.CALL.RULES.BIN and friends
########################################################################

########################################################################
# Subdirs support...
#
# The current dir is target [subdir-.]. Subdirs are created as
# subdir-DIRNAME, with a dependency on subdir-. so that the current
# dir is built before subdirs. To change the order of subdirs, use
# deps, e.g.: subdir-bar: subdir-foo
#
# The generated clean-DIRNAME and distclean-DIRNAME rules work
# differently, in that they build depth-first.
.PHONY: subdir-. subdirs
subdir-.: # implement this for the current dir if needed.
subdirs: subdir-.
########################################################################
define ShakeNMake.CALL.SUBDIR
# $1 = dir
# $2 = optional target
.PHONY: subdir-$(1)
subdir-$(1): subdir-.
	@deep=0; while [[ $$$$deep -le $$(MAKELEVEL) ]]; do echo -n $$(SUBMAKE.INDENT); deep=$$$$((deep+1)); done; \
	echo "Making subdir [$(1)]... "; \
	$(MAKE) -C $(1) --quiet --no-print-directory $(2)
clean-.: clean-$(1)
distclean-.: | distclean-$(1)
subdirs: subdir-$(1)
endef
define ShakeNMake.CALL.SUBDIRS
# $1 = dirs
$(foreach D,$(1),$(eval $(call ShakeNMake.CALL.SUBDIR,$(D))))
endef
# /subdirs
########################################################################

########################################################################
# Automatic dependencies generation for C/C++ code...
# To disable deps generation, set ShakeNMake.USE_MKDEPS=0 *before*
# including this file.
ShakeNMake.USE_MKDEPS := 1
ShakeNMake.CISH_DEPS_FILE := .make.c_deps
$(ShakeNMake.CISH_DEPS_FILE):
ifeq (,$(ShakeNMake.BINS.GCC))
ShakeNMake.USE_MKDEPS ?= 0
else
ShakeNMake.USE_MKDEPS ?= 1
endif
#$(warning ShakeNMake.USE_MKDEPS=$(ShakeNMake.USE_MKDEPS));
ifeq (1,$(ShakeNMake.USE_MKDEPS))
ShakeNMake.CISH_SOURCES ?= $(wildcard *.cpp *.c *.c++ *.h *.hpp *.h++ *.hh)
#$(warning ShakeNMake.CISH_SOURCES=$(ShakeNMake.CISH_SOURCES))

ifneq (,$(ShakeNMake.CISH_SOURCES))
ShakeNMake.BINS.MKDEP = gcc -w -E -MM $(CPPFLAGS)
#$(ShakeNMake.CISH_SOURCES): $(filter-out $(ShakeNMake.CISH_DEPS_FILE),$(MAKEFILE_LIST))
CLEAN_FILES += $(ShakeNMake.CISH_DEPS_FILE)
$(ShakeNMake.CISH_DEPS_FILE): $(MAKEFILE_LIST) $(ShakeNMake.CISH_SOURCES)
	@touch $@; test x = "x$(ShakeNMake.CISH_SOURCES)" && exit 0; \
	[ x1 = "x$(ShakeNMake.QUIET)" ] || echo "Generating dependencies..."; \
	$(ShakeNMake.BINS.MKDEP) $(ShakeNMake.CISH_SOURCES) 2>/dev/null > $@ || exit;
# perl -i -pe 's|^(.+)\.o:\s*((\w+/)*)|$$2$$1.o: $(PACKAGE.MAKEFILE) $(ShakeNMake.MAKEFILE) $$2|' $@
#    perl -i -pe 's|^(.+)\.o:\s*((\w+/)*)|$$2$$1.o: $(filter-out $(ShakeNMake.CISH_DEPS_FILE),$(MAKEFILE_LIST))\n$$2$$1.o: $$2|' $@
# gcc -E -MM strips directory parts (why, i cannot imagine), so we've got to
# hack the output.
# Example: input= src/Foo.cpp
# Generated deps:
#  Foo.o: src/Foo.cpp ...
# So we re-extract that first path and prepend it to Foo.o using the
# above perl.  We also add $(MAKEFILE_LIST) has a prereq of all src
# files because i belong to the school of tought which believes that
# changes to the arbitrary build options set in the makefiles should
# cause a rebuild.
#ifeq (,$(SRC_DIR))
#$(error Tree-specific hack in place here. See the comments and remove these lines or set SRC_DIR)
#endif
# normally we also want:
#  || $(ShakeNMake.BINS.RM) -f $@ 2>/dev/null
# because we don't want a bad generated makefile to kill the build, but gcc -E
# is returning 1 when some generated files do not yet exist.
deps: $(ShakeNMake.CISH_DEPS_FILE)

ifneq (,$(strip $(filter distclean clean,$(MAKECMDGOALS))))
# $(warning Skipping C/C++ deps generation.)
# ABSOLUTEBOGO := $(shell $(ShakeNMake.BINS.RM) -f $(ShakeNMake.CISH_DEPS_FILE))
else
#$(warning Including C/C++ deps.)
-include $(ShakeNMake.CISH_DEPS_FILE)
endif

endif
# ^^^^ ifneq(,$(ShakeNMake.CISH_SOURCES))
endif
# ^^^^ end $(ShakeNMake.USE_MKDEPS)
########################################################################

########################################################################
# Deps, Tromey's Way:
# $(call ShakeNBake.CALL.MkDep.Tromey,source-file,object-file,depend-file)
ifeq (0,1)
.dfiles.list = $(wildcard *.o.d) $(shell find src -name '*.o.d')
ifneq (,$(.dfiles.list))
# $(error include $(.dfiles.list))
include $(.dfiles.list)
endif

define ShakeNBake.CALL.MkDep.Tromey
$(ShakeNMake.BINS.GCC) -MM \
	-w	\
	-MF $(1).d         \
	-MP            \
	-MT $2         \
	$(CPPFLAGS)    \
	$(TARGET_ARCH) \
	$1
endef

CLEAN_FILES += $(wildcard *.o.d)
%.o: %.c
	@$(call ShakeNMake.CALL.SETX,"CC [$@] ..."); \
	$(COMPILE.c) -MM -MF $(@).d -MP -MT $@.d $(COMPILE.c.OPTIONS) $(OUTPUT_OPTION) $<
#	$(call ShakeNBake.CALL.MkDep.Tromey,$<,$@,$(@).d);

%.o: %.cpp
	@$(call ShakeNMake.CALL.SETX,"C++ [$@] ..."); \
	$(COMPILE.cc) -MM -MF $@.d -MP -MT $@ $(COMPILE.cc.OPTIONS) $(OUTPUT_OPTION) $<
#	$(COMPILE.cc) $(COMPILE.cc.OPTIONS) $(OUTPUT_OPTION) $<
#	$(call ShakeNBake.CALL.MkDep.Tromey,$<,$@,$(@).d);
else
%.o: %.c
	@$(call ShakeNMake.CALL.SETX,"CC [$@] ..."); \
	$(COMPILE.c) $(COMPILE.c.OPTIONS) $(OUTPUT_OPTION) $<

%.o: %.cpp
	@$(call ShakeNMake.CALL.SETX,"C++ [$@] ..."); \
	$(COMPILE.cc) $(COMPILE.cc.OPTIONS) $(OUTPUT_OPTION) $<
endif # Tromey's Way

########################################################################
# create emacs tags...
.PHONY: tags
tags:
TAGFILE := $(ShakeNMake.DIR)/TAGS
$(TAGFILE):
#ifeq ($(CURDIR),$(ShakeNMake.DIR))
ifeq (0,$(MAKELEVEL))
bin_etags := $(call ShakeNMake.CALL.FIND_FILE,etags)
ifneq (,$(bin_etags))
TAGTHESE := $(shell find $(ShakeNMake.DIR) -name '*.[ch]' -print | grep -v amalgam)
ifneq (,$(TAGTHESE))
    $(TAGTHESE):
    $(TAGFILE): $(TAGTHESE) # $(MAKEFILE_LIST)
	@echo "Creating $@..."; \
	for x in $(TAGTHESE); do echo $$x; done | $(bin_etags) -f $@ -
    tags: $(TAGFILE)
    CLEAN_FILES += $(TAGFILE)
    all: $(TAGFILE)
endif
# $(TAGTHESE)
endif
# ^^^ bin_etags
endif
# ^^^ $(MAKELEVEL)

endif
#^^^^ /ifeq (,$(ShakeNMake.MAKEFILE))
